from ovos_plugin_manager.templates.embeddings import EmbeddingsDB, TextEmbeddingsStore, FaceEmbeddingsStore, VoiceEmbeddingsStore
from ovos_plugin_manager.utils import PluginTypes


def find_embeddings_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.EMBEDDINGS)


def load_embeddings_plugin(module_name: str) -> type(EmbeddingsDB):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.EMBEDDINGS)


def find_voice_embeddings_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.VOICE_EMBEDDINGS)


def load_voice_embeddings_plugin(module_name: str) -> type(VoiceEmbeddingsStore):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.VOICE_EMBEDDINGS)


def find_face_embeddings_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.FACE_EMBEDDINGS)


def load_face_embeddings_plugin(module_name: str) -> type(FaceEmbeddingsStore):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.FACE_EMBEDDINGS)


def find_text_embeddings_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.TEXT_EMBEDDINGS)


def load_text_embeddings_plugin(module_name: str) -> type(TextEmbeddingsStore):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.TEXT_EMBEDDINGS)



"""
This is here to allow importing this module outside mycroft-core, plugins
using this import instead of mycroft can be used

The main use case is for plugins to be used across different projects
"""
import json
from abc import ABCMeta, abstractmethod
from queue import Queue
from threading import Thread, Event
from typing import List, Tuple, Optional

from ovos_config import Configuration
from ovos_utils import classproperty
from ovos_utils.log import deprecated
from ovos_utils.process_utils import RuntimeRequirements
from ovos_utils.lang import standardize_lang_tag
from ovos_plugin_manager.utils.config import get_plugin_config


class STT(metaclass=ABCMeta):
    """ STT Base class, all  STT backends derives from this one. """

    def __init__(self, config=None):
        self.config_core = Configuration()
        self._lang = None
        self._credential = None
        self._keys = None

        self.config = get_plugin_config(config, "stt")

        self.can_stream = False
        self._recognizer = None

    @classproperty
    def runtime_requirements(self):
        """ skill developers should override this if they do not require connectivity
         some examples:
         IOT plugin that controls devices via LAN could return:
            scans_on_init = True
            RuntimeRequirements(internet_before_load=False,
                                 network_before_load=scans_on_init,
                                 requires_internet=False,
                                 requires_network=True,
                                 no_internet_fallback=True,
                                 no_network_fallback=False)
         online search plugin with a local cache:
            has_cache = False
            RuntimeRequirements(internet_before_load=not has_cache,
                                 network_before_load=not has_cache,
                                 requires_internet=True,
                                 requires_network=True,
                                 no_internet_fallback=True,
                                 no_network_fallback=True)
         a fully offline plugin:
            RuntimeRequirements(internet_before_load=False,
                                 network_before_load=False,
                                 requires_internet=False,
                                 requires_network=False,
                                 no_internet_fallback=True,
                                 no_network_fallback=True)
        """
        return RuntimeRequirements()

    @property
    @deprecated("self.recognizer has been deprecated! "
                "if you need it 'from speech_recognition import Recognizer' directly", "1.0.0")
    def recognizer(self):
        # only imported here to not drag dependency
        from speech_recognition import Recognizer
        if not self._recognizer:
            self._recognizer = Recognizer()
        return self._recognizer

    @recognizer.setter
    def recognizer(self, val):
        self._recognizer = val

    @property
    def lang(self):
        return standardize_lang_tag(self._lang or \
            self.config.get("lang") or \
            Configuration().get("lang", "en-US"))

    @lang.setter
    def lang(self, val):
        # backwards compat
        self._lang = standardize_lang_tag(val)

    @property
    @deprecated("self.keys has been deprecated! "
                "implement config handling directly instead", "1.0.0")
    def keys(self):
        return self._keys or self.config_core.get("keys", {})

    @keys.setter
    def keys(self, val):
        # backwards compat
        self._keys = val

    @property
    @deprecated("self.credential has been deprecated! "
                "implement config handling directly instead", "1.0.0")
    def credential(self):
        return self._credential or self.config.get("credential", {})

    @credential.setter
    def credential(self, val):
        # backwards compat
        self._credential = val

    @staticmethod
    @deprecated("self.init_language has been deprecated! "
                "implement config handling directly instead", "1.0.0")
    def init_language(config_core):
        lang = config_core.get("lang", "en-US")
        return standardize_lang_tag(lang, macro=True)

    @abstractmethod
    def execute(self, audio, language: Optional[str] = None) -> str:
        # TODO - eventually deprecate this and make transcribe the @abstractmethod
        pass

    def transcribe(self, audio, lang: Optional[str] = None) -> List[Tuple[str, float]]:
        """transcribe audio data to a list of
        possible transcriptions and respective confidences"""
        return [(self.execute(audio, lang), 1.0)]

    @property
    def available_languages(self) -> set:
        """Return languages supported by this STT implementation in this state
        This property should be overridden by the derived class to advertise
        what languages that engine supports.
        Returns:
            set: supported languages
        """
        return set()


class TokenSTT(STT, metaclass=ABCMeta):
    @deprecated("TokenSTT is deprecated, please subclass from STT directly", "1.0.0")
    def __init__(self, config=None):
        super().__init__(config)
        self.token = self.credential.get("token")


class GoogleJsonSTT(STT, metaclass=ABCMeta):
    @deprecated("GoogleJsonSTT is deprecated, please subclass from STT directly", "1.0.0")
    def __init__(self, config=None):
        super().__init__(config)
        if not self.credential.get("json") or self.keys.get("google_cloud"):
            self.credential["json"] = self.keys["google_cloud"]
        self.json_credentials = json.dumps(self.credential.get("json"))


class BasicSTT(STT, metaclass=ABCMeta):
    @deprecated("BasicSTT is deprecated, please subclass from STT directly", "1.0.0")
    def __init__(self, config=None):
        super().__init__(config)
        self.username = str(self.credential.get("username"))
        self.password = str(self.credential.get("password"))


class KeySTT(STT, metaclass=ABCMeta):

    @deprecated("KeySTT is deprecated, please subclass from STT directly", "1.0.0")
    def __init__(self, config=None):
        super().__init__(config)
        self.id = str(self.credential.get("client_id"))
        self.key = str(self.credential.get("client_key"))


class StreamThread(Thread, metaclass=ABCMeta):
    """
        ABC class to be used with StreamingSTT class implementations.
    """

    def __init__(self, queue, language):
        super().__init__()
        self.language = standardize_lang_tag(language)
        self.queue = queue
        self.text = None

    def _get_data(self):
        while True:
            d = self.queue.get()
            if d is None:
                break
            yield d
            self.queue.task_done()

    def run(self):
        return self.handle_audio_stream(self._get_data(), self.language)

    def finalize(self):
        """ return final transcription """
        return self.text

    @abstractmethod
    def handle_audio_stream(self, audio, language):
        pass


class StreamingSTT(STT, metaclass=ABCMeta):
    """
        ABC class for threaded streaming STT implementations.
    """

    def __init__(self, config=None):
        super().__init__(config)
        self.stream = None
        self.can_stream = True
        self.transcript_ready = Event()

    def stream_start(self, language=None):
        self.stream_stop()
        self.queue = Queue()
        self.stream = self.create_streaming_thread()
        self.stream.language = standardize_lang_tag(language or self.lang)
        self.transcript_ready.clear()
        self.stream.start()

    def stream_data(self, data):
        self.queue.put(data)

    def stream_stop(self):
        if self.stream is not None:
            self.queue.put(None)
            text = self.stream.finalize()
            self.stream.join()
            self.stream = None
            self.queue = None
            self.transcript_ready.set()
            return text
        return None

    def execute(self, audio: Optional = None,
                language: Optional[str] = None):
        return self.stream_stop()

    def transcribe(self, audio: Optional = None,
                   lang: Optional[str] = None) -> List[Tuple[str, float]]:
        """transcribe audio data to a list of
        possible transcriptions and respective confidences"""
        return [(self.execute(audio, lang), 1.0)]

    @abstractmethod
    def create_streaming_thread(self):
        pass



import hashlib
import json
import os
from os.path import join, isdir
import shutil
from pathlib import Path
from stat import S_ISREG, ST_MTIME, ST_MODE, ST_SIZE

from ovos_config.locations import get_xdg_cache_save_path
from ovos_utils.file_utils import get_cache_directory as get_tmp_cache_dir
from ovos_utils.log import LOG


def hash_sentence(sentence: str):
    """Convert the sentence into a hash value used for the file name

    Args:
        sentence: The sentence to be cached
    """
    encoded_sentence = sentence.encode("utf-8", "ignore")
    sentence_hash = hashlib.md5(encoded_sentence).hexdigest()
    return sentence_hash


def hash_from_path(path: Path) -> str:
    """Returns hash from a given path.

    Simply removes extension and folder structure leaving the hash.

    NOTE: this does not do any hashing at all and naming is misleading
          however we keep the method around for backwards compat imports
          this is exclusively for usage with cached TTS files

    Args:
        path: path to get hash from

    Returns:
        Hash reference for file.
    """
    # NOTE: this does not do any hashing at all and naming is misleading
    # however we keep the method around for backwards compat imports
    # this is assumed to be used only to load cached TTS which is already named with an hash
    return path.with_suffix('').name


def mb_to_bytes(size):
    """Takes a size in MB and returns the number of bytes.

    Args:
        size(int/float): size in Mega Bytes

    Returns:
        (int/float) size in bytes
    """
    return size * 1024 * 1024


def _get_cache_entries(directory):
    """Get information tuple for all regular files in directory.

    Args:
        directory (str): path to directory to check

    Returns:
        (tuple) (modification time, size, filepath)
    """
    entries = (os.path.join(directory, fn) for fn in os.listdir(directory))
    entries = ((os.stat(path), path) for path in entries)

    # leave only regular files, insert modification date
    return ((stat[ST_MTIME], stat[ST_SIZE], path)
            for stat, path in entries if S_ISREG(stat[ST_MODE]))


def _delete_oldest(entries, bytes_needed):
    """Delete files with oldest modification date until space is freed.

    Args:
        entries (tuple): file + file stats tuple
        bytes_needed (int): disk space that needs to be freed

    Returns:
        (list) all removed paths
    """
    deleted_files = []
    space_freed = 0
    for moddate, fsize, path in sorted(entries):
        try:
            os.remove(path)
            space_freed += fsize
            deleted_files.append(path)
        except Exception:
            pass

        if space_freed > bytes_needed:
            break  # deleted enough!

    return deleted_files


def curate_cache(directory, min_free_percent=5.0, min_free_disk=50):
    """Clear out the directory if needed.

    The curation will only occur if both the precentage and actual disk space
    is below the limit. This assumes all the files in the directory can be
    deleted as freely.

    Args:
        directory (str): directory path that holds cached files
        min_free_percent (float): percentage (0.0-100.0) of drive to keep free,
                                  default is 5% if not specified.
        min_free_disk (float): minimum allowed disk space in MB, default
                               value is 50 MB if not specified.
    """
    # Simpleminded implementation -- keep a certain percentage of the
    # disk available.
    # TODO: Would be easy to add more options, like whitelisted files, etc.
    deleted_files = []

    if not isdir(directory):
        raise NotADirectoryError(directory)

    # Get the disk usage statistics bout the given path
    space = shutil.disk_usage(directory)

    percent_free = space.free * 100 / space.total

    min_free_disk = mb_to_bytes(min_free_disk)

    if percent_free < min_free_percent and space.free < min_free_disk:
        LOG.info('Low diskspace detected, cleaning cache')
        # calculate how many bytes we need to delete
        bytes_needed = (min_free_percent - percent_free) / 100.0 * space.total
        bytes_needed = int(bytes_needed + 1.0)

        # get all entries in the directory w/ stats
        entries = _get_cache_entries(directory)
        # delete as many as needed starting with the oldest
        deleted_files = _delete_oldest(entries, bytes_needed)

    return deleted_files


class AudioFile:
    def __init__(self, cache_dir: Path, sentence_hash: str, file_type: str):
        self.name = f"{sentence_hash}.{file_type}"
        if isinstance(cache_dir, str):
            cache_dir = Path(cache_dir)
        self.path = cache_dir.joinpath(self.name)
        self.audio_data = None

    def save(self, audio: bytes):
        """Write a TTS cache file containing the audio to be spoken.
        Args:
            audio: TTS inference of a sentence
        """
        try:
            self.audio_data = audio
            with open(self.path, "wb") as audio_file:
                audio_file.write(audio)
        except Exception:
            LOG.exception(f"Failed to write {self.name} to cache")

    def load(self):
        """Load audio data from cached file."""
        if self.path.exists():
            try:
                with open(self.path, "rb") as audio:
                    self.audio_data = audio.read()
            except Exception:
                LOG.exception(f"Failed to read {self.name} audio data from cache")
        return self.audio_data

    def exists(self):
        return self.path.exists()

    def __str__(self):
        return str(self.path)


class PhonemeFile:
    def __init__(self, cache_dir: Path, sentence_hash: str):
        self.name = f"{sentence_hash}.pho"
        if isinstance(cache_dir, str):
            cache_dir = Path(cache_dir)
        self.path = cache_dir.joinpath(self.name)
        self.phonemes = None

    def load(self):
        """Load phonemes from cache file."""
        if self.path.exists():
            try:
                with open(self.path) as phoneme_file:
                    phonemes = phoneme_file.read().strip()
                self.phonemes = json.loads(phonemes)
            except Exception:
                LOG.exception(f"Failed to read {self.name} phonemes from cache")
        return self.phonemes

    def save(self, phonemes):
        """Write a TTS cache file containing the phoneme to be displayed.
        Args:
            phonemes: instructions for how to make the mouth on a device move
        """
        self.phonemes = phonemes
        try:
            rec = json.dumps(phonemes)
            with open(self.path, "w") as phoneme_file:
                phoneme_file.write(rec)
        except Exception:
            LOG.error(f"Failed to write {self.name} to cache")

    def exists(self):
        return self.path.exists()

    def __str__(self):
        return str(self.path)


class TextToSpeechCache:
    """Class for all persistent and temporary caching operations."""

    def __init__(self, tts_config, tts_name, audio_file_type):
        self.config = tts_config
        self.tts_name = tts_name
        self.audio_file_type = audio_file_type

        persistent_cache = self.config.get("preloaded_cache") or \
                           join(get_xdg_cache_save_path(), tts_name)
        tmp_cache = get_tmp_cache_dir(f"tts/{tts_name}")
        os.makedirs(tmp_cache, exist_ok=True)
        os.makedirs(persistent_cache, exist_ok=True)

        self.persistent_cache_dir = Path(persistent_cache)
        self.temporary_cache_dir = Path(tmp_cache)
        self.cached_sentences = {}
        # curate cache if disk usage is above min %
        self.min_free_percent = self.config.get("min_free_percent", 75)
        # save to permanent cache settings
        self.persist = self.config.get("persist_cache", False)
        # only persist if utterance is spoken >= N times
        self.persist_thresh = self.config.get("persist_thresh", 1)
        self._sentence_count = {}
        self.load_persistent_cache()

    def __contains__(self, sha):
        """The cache contains a SHA if it knows of it and it exists on disk."""
        if sha not in self.cached_sentences:
            return False  # Doesn't know of it
        else:
            # Audio file must exist, phonemes are optional.
            audio, phonemes = self.cached_sentences[sha]
            return (audio.exists() and
                    (phonemes is None or phonemes.exists()))

    def load_persistent_cache(self):
        """There may be files pre-loaded in the persistent cache directory
        prior to run time, such as pre-recorded audio files.
        """
        if self.persistent_cache_dir is not None:
            self._load_existing_audio_files()
            self._load_existing_phoneme_files()
            LOG.info("Persistent TTS cache files loaded successfully.")

    def _load_existing_audio_files(self):
        """Find the TTS audio files already in the persistent cache."""
        glob_pattern = "*." + self.audio_file_type
        for file_path in self.persistent_cache_dir.glob(glob_pattern):
            sentence_hash = file_path.name.split(".")[0]
            audio_file = self.define_audio_file(sentence_hash, persistent=True)
            self.cached_sentences[sentence_hash] = audio_file, None

    def _load_existing_phoneme_files(self):
        """Find the TTS phoneme files already in the persistent cache.
        A phoneme file is no good without an audio file to pair it with.  If
        no audio file matches, do not load the phoneme.
        """
        for file_path in self.persistent_cache_dir.glob("*.pho"):
            sentence_hash = file_path.name.split(".")[0]
            cached_sentence = self.cached_sentences.get(sentence_hash)
            if cached_sentence is not None:
                audio_file = cached_sentence[0]
                phoneme_file = self.define_phoneme_file(sentence_hash, persistent=True)
                self.cached_sentences[sentence_hash] = audio_file, phoneme_file

    def clear(self):
        """Remove all files from the temporary cache."""
        for cache_file_path in self.temporary_cache_dir.iterdir():
            if cache_file_path.is_dir():
                for sub_path in cache_file_path.iterdir():
                    if sub_path.is_file():
                        sub_path.unlink()
            elif cache_file_path.is_file():
                cache_file_path.unlink()

    def curate(self):
        """Remove cache data if disk space is running low."""
        try:
            files_removed = curate_cache(str(self.temporary_cache_dir),
                                         min_free_percent=self.min_free_percent)
        except NotADirectoryError:
            LOG.info(f"Nothing to curate")
            return
        hashes = set([hash_from_path(Path(path)) for path in files_removed])
        for sentence_hash in hashes:
            if sentence_hash in self.cached_sentences:
                self.cached_sentences.pop(sentence_hash)

    def define_audio_file(self, sentence_hash: str, persistent=False) -> AudioFile:
        """Build an instance of an object representing an audio file."""
        if persistent or self._should_persist(sentence_hash):
            audio_file = AudioFile(
                self.persistent_cache_dir, sentence_hash, self.audio_file_type
            )
        else:
            audio_file = AudioFile(
                self.temporary_cache_dir, sentence_hash, self.audio_file_type
            )
        return audio_file

    def define_phoneme_file(self, sentence_hash: str, persistent=False) -> PhonemeFile:
        """Build an instance of an object representing an phoneme file."""
        if persistent or self._should_persist(sentence_hash):
            phoneme_file = PhonemeFile(self.persistent_cache_dir, sentence_hash)
        else:
            phoneme_file = PhonemeFile(self.temporary_cache_dir, sentence_hash)
        return phoneme_file

    def _should_persist(self, sentence_hash: str):
        if self.persist:
            if sentence_hash not in self._sentence_count:
                self._sentence_count[sentence_hash] = 0
            self._sentence_count[sentence_hash] += 1
            if self._sentence_count[sentence_hash] >= self.persist_thresh:
                return True
        return False



from ovos_plugin_manager.utils import PluginTypes
from ovos_plugin_manager.templates.ocp import OCPStreamExtractor
from ovos_utils.log import LOG
from functools import lru_cache

from ovos_plugin_manager.utils import find_plugins

try:
    from ovos_plugin_manager.templates.media import AudioPlayerBackend, VideoPlayerBackend, WebPlayerBackend
except ImportError:
    LOG.warning("Please install ovos-utils~=0.1 for `AudioPlayerBackend`, "
                "`VideoPlayerBackend`, and `WebPlayerBackend` imports.")


def find_ocp_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    return find_plugins(PluginTypes.STREAM_EXTRACTOR)


def find_ocp_audio_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    return find_plugins(PluginTypes.AUDIO_PLAYER)


def find_ocp_video_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    return find_plugins(PluginTypes.VIDEO_PLAYER)


def find_ocp_web_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    return find_plugins(PluginTypes.WEB_PLAYER)


class StreamHandler:
    def __init__(self):
        self.extractors = {}
        self.load()

    @property
    def supported_seis(self):
        """
        skills may return results requesting a specific extractor to be used

        plugins should report a StreamExtractorIds (sei) that identifies it can handle certain kinds of requests

        any streams of the format "{sei}//{uri}" can be handled by this plugin
        """
        seis = []
        for extractor in self.extractors.values():
            seis += extractor.supported_seis
        return seis

    def load(self):
        for plugin, clazz in find_ocp_plugins().items():
            try:
                self.extractors[plugin] = clazz()
                LOG.info(f"Loaded OCP plugin: {plugin}")
            except:
                LOG.error(f"Failed to load {plugin}")
                continue

    def _get_sei_plugs(self, uri):
        return [plug for plug in self.extractors.values()
                if any((uri.startswith(f"{sei}//") for sei in plug.supported_seis))]

    def _extract_from_sei(self, uri, video=True):
        # attempt to use a dedicated stream extractor if requested
        for plug in self._get_sei_plugs(uri):
            try:
                return plug.extract_stream(uri, video)
            except Exception as e:
                LOG.exception(f"error extracting stream with {plug}")

    def _extract_from_url(self, uri, video=True):
        for plug in self.extractors.values():
            try:
                if plug.validate_uri(uri):
                    return plug.extract_stream(uri, video)
            except Exception as e:
                LOG.exception(f"error extracting stream with {plug}")

    def extract_stream(self, uri, video=True):
        meta = {}

        # attempt to use a dedicated stream extractor if requested
        while len(self._get_sei_plugs(uri)):  # support chained extractions, where one plugin calls another
            meta = self._extract_from_sei(uri, video) or {}
            if meta.get("uri"):
                uri = meta["uri"]
            else:
                break

        # let plugins parse the raw url and see if they want to handle it
        meta = self._extract_from_url(uri, video) or meta

        # no extractor available, return raw url
        return meta or {"uri": uri}


@lru_cache()  # to avoid loading StreamHandler more than once
def load_stream_extractors():
    return StreamHandler()


def available_extractors():
    xtract = load_stream_extractors()
    return ["/", "http:", "https:", "file:"] + \
        [f"{sei}//" for sei in xtract.supported_seis]



import abc
from collections import namedtuple
from dataclasses import dataclass
from typing import Optional, Dict, List, Union

from ovos_bus_client.client import MessageBusClient
from ovos_bus_client.message import Message
from ovos_utils.fakebus import FakeBus

# LEGACY: Intent match response tuple, ovos-core~=0.2 expects PipelinePlugin to return this data structure
# intent_service: Name of the service that matched the intent
# intent_type: intent name (used to call intent handler over the message bus)
# intent_data: data provided by the intent match
# skill_id: the skill this handler belongs to
# TODO - deprecated
IntentMatch = namedtuple('IntentMatch',
                         ['intent_service', 'intent_type',
                          'intent_data', 'skill_id', 'utterance']
                         )


@dataclass
class IntentHandlerMatch:
    """
    Represents an intent handler match result, expected by ovos-core plugins.

    Attributes:
        match_type (str): Name of the service that matched the intent.
        match_data (Optional[Dict]): Additional data provided by the intent match.
        skill_id (Optional[str]): The skill this handler belongs to.
        utterance (Optional[str]): The original utterance triggering the intent.
    """
    match_type: str
    match_data: Optional[Dict] = None
    skill_id: Optional[str] = None
    utterance: Optional[str] = None


@dataclass
class PipelineMatch(IntentHandlerMatch):
    """
    Represents a match in a pipeline that does not trigger an intent message directly.

    Attributes:
        match_type (bool): Indicates if the utterance was matched (compatibility only).
        handled (bool): Whether the match has already handled the utterance.
        match_data (Optional[Dict]): Data provided by the intent match.
        skill_id (Optional[str]): The skill this handler belongs to.
        utterance (Optional[str]): The original utterance triggering the match.
    """
    handled: bool = True
    match_data: Optional[Dict] = None
    skill_id: Optional[str] = None
    utterance: Optional[str] = None
    match_type: bool = True  # compat


class PipelinePlugin:
    """
    Base class for intent matching pipeline plugins. Mainly useful for typing

    Attributes:
        config (Dict): Configuration for the plugin.
    """

    def __init__(self, config: Optional[Dict] = None):
        self.config = config or {}


class ConfidenceMatcherPipeline(PipelinePlugin):
    """
    Base class for plugins that match utterances with confidence levels,
    but do not directly trigger actions.

    Example plugins: adapt, padatious.

    Attributes:
        bus (Union[MessageBusClient, FakeBus]): The message bus client for communication.
    """

    def __init__(self, bus: Optional[Union[MessageBusClient, FakeBus]] = None,
                 config: Optional[Dict] = None):
        self.bus = bus or FakeBus()
        super().__init__(config=config)

    @abc.abstractmethod
    def match_high(self, utterances: List[str], lang: str, message: Message) -> Optional[IntentHandlerMatch]:
        """
        Match an utterance with high confidence.

        Args:
            utterances (List[str]): List of utterances to match.
            lang (str): The language of the utterances.
            message (Message): The message containing the utterance.

        Returns:
            Optional[IntentHandlerMatch]: The match result or None if no match is found.
        """

    @abc.abstractmethod
    def match_medium(self, utterances: List[str], lang: str, message: Message) -> Optional[IntentHandlerMatch]:
        """
        Match an utterance with medium confidence.

        Args:
            utterances (List[str]): List of utterances to match.
            lang (str): The language of the utterances.
            message (Message): The message containing the utterance.

        Returns:
            Optional[IntentHandlerMatch]: The match result or None if no match is found.
        """

    @abc.abstractmethod
    def match_low(self, utterances: List[str], lang: str, message: Message) -> Optional[IntentHandlerMatch]:
        """
        Match an utterance with low confidence.

        Args:
            utterances (List[str]): List of utterances to match.
            lang (str): The language of the utterances.
            message (Message): The message containing the utterance.

        Returns:
            Optional[IntentHandlerMatch]: The match result or None if no match is found.
        """


class PipelineStageMatcher(PipelinePlugin):
    """
    Base class for plugins that consume an utterance during matching,
    aborting subsequent pipeline stages if a match is found.

    WARNING: has side effects when match is used

    these plugins will consume an utterance during the match process,
    it is not known if this component will match without going through the match process

    Example plugins: converse, common_query.

    Attributes:
        bus (Union[MessageBusClient, FakeBus]): The message bus client for communication.
    """

    def __init__(self, bus: Optional[Union[MessageBusClient, FakeBus]] = None,
                 config: Optional[Dict] = None):
        self.bus = bus or FakeBus()
        super().__init__(config=config)

    @abc.abstractmethod
    def match(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """
        Match an utterance, potentially aborting further stages in the pipeline.

        Args:
            utterances (List[str]): List of utterances to match.
            lang (str): The language of the utterances.
            message (Message): The message containing the utterance.

        Returns:
            Optional[PipelineMatch]: The match result or None if no match is found.
        """
        pass


class PipelineStageConfidenceMatcher(PipelineStageMatcher, ConfidenceMatcherPipeline):
    """
    A specialized matcher that consumes utterances during the matching process
    and supports confidence levels. It aborts further pipeline stages if a match is found.

    Example plugins: fallback, stop.

    Attributes:
        bus (Union[MessageBusClient, FakeBus]): The message bus client for communication.
    """

    def __init__(self, bus: Optional[Union[MessageBusClient, FakeBus]] = None,
                 config: Optional[Dict] = None):
        super().__init__(bus=bus, config=config)

    def match(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """
        Match an utterance using high confidence, with no specific match level defined.

        Args:
            utterances (List[str]): List of utterances to match.
            lang (str): The language of the utterances.
            message (Message): The message containing the utterance.

        Returns:
            Optional[PipelineMatch]: The match result or None if no match is found.
        """
        return self.match_high(utterances, lang, message)

    @abc.abstractmethod
    def match_high(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """
        Match an utterance with high confidence.

        Args:
            utterances (List[str]): List of utterances to match.
            lang (str): The language of the utterances.
            message (Message): The message containing the utterance.

        Returns:
            Optional[PipelineMatch]: The match result or None if no match is found.
        """

    @abc.abstractmethod
    def match_medium(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """
        Match an utterance with medium confidence.

        Args:
            utterances (List[str]): List of utterances to match.
            lang (str): The language of the utterances.
            message (Message): The message containing the utterance.

        Returns:
            Optional[PipelineMatch]: The match result or None if no match is found.
        """

    @abc.abstractmethod
    def match_low(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """
        Match an utterance with low confidence.

        Args:
            utterances (List[str]): List of utterances to match.
            lang (str): The language of the utterances.
            message (Message): The message containing the utterance.

        Returns:
            Optional[PipelineMatch]: The match result or None if no match is found.
        """



from typing import Optional, Dict, Any

from ovos_bus_client import Message
from ovos_bus_client import MessageBusClient
from ovos_bus_client.apis.gui import GUIInterface
from ovos_config import Configuration
from ovos_utils.log import LOG

try:
    from ovos_gui.homescreen import HomescreenManager
except ImportError as _exc:

    class HomescreenManager:
        """Fallback class when ovos-gui is not installed.

        Raises the original ImportError when instantiated to
        provide clear error messaging while still allowing type hints to work.
        """

        def __init__(self, *args, **kwargs):
            LOG.error("you seem to be running GUIExtensions without ovos-gui installed...")
            # raise the original ImportError
            raise _exc


class GUIExtension:
    """ GUI Extension base class

    These plugins are responsible for managing the GUI behaviours
    for specific platforms such as homescreen handling

    only 1 GUIExtension is loaded at any time by ovos-gui service

    Args:
        bus: MessageBus instance
        gui: GUI instance
        preload_gui (bool): load GUI skills even if gui client not connected
        permanent (bool): disable unloading of GUI skills on gui client disconnections
    """

    def __init__(self, config: Dict[str, Any],
                 bus: Optional[MessageBusClient] = None,
                 gui: Optional[GUIInterface] = None,
                 preload_gui: bool = False,
                 permanent: bool = False):

        if not bus:
            bus = MessageBusClient()
            bus.run_in_thread()
            bus.connected_event.wait()
        self.bus: MessageBusClient = bus
        self.gui: GUIInterface = gui or GUIInterface("ovos.shell", bus=self.bus,
                                                     config=Configuration().get("gui", {}))
        self.preload_gui = preload_gui
        self.permanent = permanent
        self.config = config
        self.homescreen_manager: Optional[HomescreenManager] = None
        self.register_bus_events()

    def register_bus_events(self):
        self.bus.on("mycroft.gui.screen.close", self.handle_remove_namespace)

    def bind_homescreen(self, homescreen: Optional[HomescreenManager] = None):
        if self.config.get("homescreen_supported", False):
            if not homescreen:
                LOG.debug("Loading HomescreenManager")
                homescreen = HomescreenManager(self.bus)
                homescreen.daemon = True
                homescreen.start()
            self.homescreen_manager = homescreen
        else:
            LOG.info("Homescreen support not configured")

    def handle_remove_namespace(self, message: Message):
        get_skill_namespace = message.data.get("skill_id", "")
        if get_skill_namespace:
            self.bus.emit(Message("gui.clear.namespace",
                                  {"__from": get_skill_namespace}))
